#if	!defined(FIRST32PM) && !defined(FIRST32ELF)
#define	FIRST32RM		/* Real mode interface */
#else
#undef	FIRST32RM
#endif

#include	"linux-asm-string.h"
#ifdef	FIRST32RM
#include	"linux-asm-io.h"
#endif	/* FIRST32RM */
#include	"etherboot.h"

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

/*

Memory layout assumed by mknbi and this program

0x07C00-0x07FFF   0.5 kB	floppy boot sector if loaded from floppy
0x0F???-0x0FFFF     ? kB	large Etherboot data buffers (some drivers)
0x10000-0x8FFFF 512.0 kB	kernel (from tagged image)
0x90000-0x901FF	  0.5 kB	Linux floppy boot sector (from tagged image)
0x90200-0x921FF	  8.0 kB	kernel setup (from tagged image)
0x92200-0x931FF	  4.0 kB	this program (from tagged image)
0x93200-0x933FF	  0.5 kB	tagged image header
0x93400-0x935FF	  0.5 kB	kernel parameters (from tagged image)
0x93600-0x93BFF   1.5 kB	(free)
0x93C00-0x93FFF	  1.0 kB	bootp record (only if BOOTP_DATA_AT_0x93C00)
0x94000-0x9FFFF	 48.0 kB	Etherboot (top few kB may be used by BIOS)
				Normally Etherboot starts at 0x94000
0x100000-			kernel (if bzImage) (from tagged image)
after bzImage kernel		ramdisk (optional) (from tagged image)
				moved to below top of memory by this program

*/

#define		PARAMSIZE	512

extern void printf(const char *, ...);
extern int sprintf(char *, const char *, ...);
extern void console_putc(int);
extern unsigned long memsize(void);
extern void xstart(unsigned long);

#ifdef	FIRST32ELF
static Elf32_Phdr	*seg[S_END] = { 0 };
#else
static struct segment	*seg[S_END] = { 0 };
#endif
static char 		*ip, *op;
static short		*vgamode;
static struct bootp_t	*bp;
static unsigned char	*vendortags;

/* Only for real mode call */
#ifdef	FIRST32RM
extern unsigned long currticks(void);

#define K_RDWR		0x60		/* keyboard data & cmds (read/write) */
#define K_STATUS	0x64		/* keyboard status */
#define K_CMD		0x64		/* keybd ctlr command (write-only) */

#define K_OBUF_FUL	0x01		/* output buffer full */
#define K_IBUF_FUL	0x02		/* input buffer full */

#define KC_CMD_WIN	0xd0		/* read  output port */
#define KC_CMD_WOUT	0xd1		/* write output port */
#define KB_SET_A20	0xdf		/* enable A20,
					   enable output buffer full interrupt
					   enable data line
					   disable clock line */
#define KB_UNSET_A20	0xdd		/* enable A20,
					   enable output buffer full interrupt
					   enable data line
					   disable clock line */
#ifndef	IBM_L40
static void empty_8042(void)
{
	unsigned long time;
	char st;

	time = currticks() + 18;	/* max wait of 1 second */
	while ((((st = inb(K_CMD)) & K_OBUF_FUL) ||
	       (st & K_IBUF_FUL)) &&
	       currticks() < time)
		inb(K_RDWR);
}
#endif	IBM_L40

/*
 * Gate A20 for high memory
 */
void gateA20_set(void)
{
#ifdef	IBM_L40
	outb(0x2, 0x92);
#else	/* IBM_L40 */
	empty_8042();
	outb(KC_CMD_WOUT, K_CMD);
	empty_8042();
	outb(KB_SET_A20, K_RDWR);
	empty_8042();
#endif	/* IBM_L40 */
}
#endif		/* FIRST32RM */

void putchar(int c)
{
	if (c == '\n')
		console_putc('\r');
	console_putc(c);
}

static inline void quit(void)
{
	printf("Bad argument\n");
	exit(0);
}

static void nomem(void)
{
	printf("Out of parameter space\n");
	exit(0);
}

static inline void checkvendor(void)
{
	union {
		unsigned long	l;
		unsigned char	c[4];
	} u;

	memcpy(u.c, vendortags, sizeof(u));
	if (u.l == RFC_1048 || u.l == VEND_CMU || u.l == VEND_STAN)
		vendortags += 4;
	else
		vendortags = 0;
}

#ifdef	FIRST32ELF
static inline void locate_segs(union infoblock *header)
{
	int		i;
	Elf32_Phdr	*s;

	s = (Elf32_Phdr *)((char *)header + header->ehdr.e_phoff);
	for (i = 0; i < header->ehdr.e_phnum; i++, s++) {
		seg[i] = s;
#if	DEBUG > 1
		printf("%d %#X\n", i, s->p_paddr);
#endif
	}
}

#else

static inline void locate_segs(union infoblock *header)
{
	int		i;
	struct segment	*s;

	s = (struct segment *)((char *)header + sizeof(struct imgheader)
		+ ((header->img.length & 0xF0) >> 2));
	for (i = 0; i < S_END; i++, s++) {
		seg[i] = s;
#if	DEBUG > 1
		printf("%d %#X\n", i, s->p_paddr);
#endif
		if (s->flags & F_FINAL)
			break;
		s = (struct segment *)((char *)s + ((s->lengths & 0xF0) >> 2));
	}
}
#endif	/* !FIRST32ELF */

static unsigned char *gettag(int tag)
{
	unsigned char		*p;
	unsigned char		c;
	static unsigned char	emptytag[] = { 0, 0 };

	if (vendortags == 0)
		return (emptytag);
	for (p = vendortags; (c = *p) != RFC1533_END; ) {
		if (c == RFC1533_PAD)
			p++;
		else if (c == tag)
			return (p + 1);
		else
			p += p[1] + 2;
	}
	return (emptytag);
}

static void outtag(unsigned char *value)
{
	int			len;

	len = *value++;
	if (op + len > ip)
		nomem();
	while (len-- > 0)
		*op++ = *value++;
}

/* Return 1 if s2 is a prefix of s1 */
static int strprefix(const char *s1, const char *s2)
{
	while (*s1 != '\0' && *s2 != '\0' && *s1++ == *s2++)
		;
	/* Have we reached the end of s2? */
	return (*s2 == '\0');
}

enum keyword { K_VGA, K_NFSROOT, K_IP };

static inline int match_keyword(const char *start)
{
	if (strprefix(start, "vga"))
		return (K_VGA);
	if (strprefix(start, "nfsroot"))
		return (K_NFSROOT);
	if (strprefix(start, "ip"))
		return (K_IP);
	return (-1);
}

#define	isws(c)	((c) == ' ' || (c) == '\t')

static inline int copy_and_match(void)
{
	int			c;
	char			*start;

	start = ip;
	/* Stop copying at = if it exists */
	while ((c = *ip) != '\0' && !isws(c) && c != '=') {
		*op++ = *ip++;
	}
	if (c == '=') {
		ip++;
		*op++ = '=';
		return (match_keyword(start));
	}
	return (-1);
}

static int gethex(const char *p)
{
	int			value = 0;

	for (;;) {
		int c = *p++;
		if (c >= '0' && c <= '9')
			c -= '0';
		else if (c >= 'a' && c <= 'f')
			c -= 'a' - 10;
		else if (c >= 'A' && c <= 'F')
			c -= 'A' - 10;
		else
			break;
		value <<= 4;
		value |= c;
	}
	return (value);
}

static int getdec(const char *p)
{
	int			value = 0, sign = 0;

	if (*p == '-') {
		sign = 1;
		p++;
	}
	for (;;) {
		int c = *p++;
		if (c >= '0' && c <= '9')
			c -= '0';
		else
			break;
		value *= 10;
		value += c;
	}
	return (sign ? -value : value);
}

static void copy_nonws(void)
{
	int			c;

	/* Copy up to next whitespace */
	while ((c = *ip) != '\0' && !isws(c))
		*op++ = *ip++;
}

static void discard_arg(void)
{
	int			c;

	/* Discard up to next whitespace */
	while ((c = *ip) != '\0' && !isws(c))
		ip++;
}

static int outip(const char *p)
{
	long		ip;

	if (*p == 0)
		return (0);
	memcpy(&ip, p + 1, sizeof(ip));
	return sprintf(op, "%I", ip);
}

static inline void subst_value(int kwindex)
{
	int			c;
	char			*p;

	if (kwindex == K_VGA) {
		/* backup over "vga=" */
		op -= sizeof("vga=") - 1;
		if (strprefix(ip, "ask"))
			c = -3;
		else if (strprefix(ip, "extended"))
			c = -2;
		else if (strprefix(ip, "normal"))
			c = -1;
		else if (strprefix(ip, "0x"))
			c = gethex(ip+2);
		else	/* assume decimal mode number */
			c = getdec(ip);
		*vgamode = c;
		discard_arg();
	} else if (kwindex == K_NFSROOT && strprefix(ip, "rom") &&
		(ip[3] == '\0' || isws(ip[3]))) {
		outtag(gettag(RFC1533_ROOTPATH));
		discard_arg();
	} else if (kwindex == K_IP && strprefix(ip, "rom") &&
		(ip[3] == '\0' || isws(ip[3]))) {
		long		ip;
		op += sprintf(op, "%I:%I:", bp->bp_yiaddr, bp->bp_siaddr);
		p = gettag(RFC1533_GATEWAY);
		op += outip(p);
		*op++ = ':';
		p = gettag(RFC1533_NETMASK);
		op += outip(p);
		*op++ = ':';
		outtag(gettag(RFC1533_HOSTNAME));
		p = gettag(RFC1533_VENDOR_ETHDEV);
		if (*p)
			*op++ = ':';
		outtag(p);
		discard_arg();
	} else
		copy_nonws();
}

static inline int skipws(void)
{
	int			c;

	while ((c = *ip) != '\0' && isws(c))
		ip++;
	return (c);
}

/*
 *	The parameters are copied from the input area to the output
 *	area, looking out for keyword=value pairs while doing so.
 *	If a possible keyword is found, indicated by an =,
 *	it is matched against a small list.
 *	If it matches none of the keywords on the list,
 *	the value is copied unchanged.
 *	If it matches a keyword, then the appropriate substitutions
 *	are made.
 *	While doing the substitution, a check is made that the output
 *	pointer doesn't overrun the input pointer. This is the only
 *	place it could happen, as the substitution may be longer than
 *	the original.
 */
static inline void process_params(void)
{
	int			i;

	while (skipws() != '\0') {
		if ((i = copy_and_match()) >= 0)
			subst_value(i);
		else
			copy_nonws();
		*op++ = ' ';
	}
	/* There may be a space after the last arg, probably does not matter
           but this is a reminder */
	*op = '\0';
}

#ifdef	FIRST32RM
int first(union infoblock *header, struct bootp_t *bootp)
#else
int first(struct ebinfo *eb, union infoblock *header, struct bootp_t *bootp)
#endif
{
	int			i;
	char			*p, *q, *params;
	struct bootblock	*boot;
	struct setupblock	*setup;
	unsigned long		top_of_mem;
	union {
		unsigned long	l;
		unsigned char	c[4];
	} u;

	printf(VERSION "/" __FILE__
#ifdef	FIRST32PM
		" (PM)"
#endif
#ifdef	FIRST32ELF
		" (ELF)"
#endif
		" (GPL)\n");
#if	DEBUG > 1
	printf("%#X %#X\n", header, bootp);
#endif
	/* Sanity checks */
#ifdef	FIRST32ELF
	if (header->img.magic != ELF_MAGIC)
#else
	if (header->img.magic != TAG_MAGIC)
#endif
		quit();
#ifdef	FIRST32RM
	gateA20_set();		/* So we can access extended memory */
#endif
	printf("%dk total memory\n", (top_of_mem = memsize()));
	bp = bootp;
	vendortags = (unsigned char *)bootp->bp_vend;
	checkvendor();
	locate_segs(header);
	/* Locate boot block */
	boot = (struct bootblock *)seg[S_BOOT]->p_paddr;
	/* Adjust boot block pointers to point to command line */
	boot->cl_magic = CL_MAGIC;
	boot->cl_offset = (params = (char *)seg[S_PARAMS]->p_paddr) - ((char *)boot);
	/* Point to word to alter if vga=... specified */
	vgamode = &boot->vgamode;
	/* Locate setup block */
	setup = (struct setupblock *)seg[S_SETUP]->p_paddr;
	/* Adjust loader type byte */
	setup->su_type = SU_MY_LOADER_TYPE;
	p = params + (i = strlen(params) + 1);
	/* Append T129 if present */
	q = gettag(RFC1533_VENDOR_MAGIC);
	/* Check T128 present and correct */
	if (*q == 6 && (memcpy(u.c, q + 1, sizeof(u)), u.l == VEND_EB)) {
		q = gettag(RFC1533_VENDOR_ADDPARM);
		i = PARAMSIZE - 1 - i;		/*  +1 for SPACE */
		if (i > *q)			/* enough space? */
			i = *q;
		q++;
		if (i > 0)
			p[-1] = ' ';		/* NUL -> space */
		while (i-- > 0)
			*p++ = *q++;
		*p++ = '\0';			/* position past NUL */
	}
	/* Move parameters to end of parameter area,
	   tail first to avoid overwriting */
	q = params + PARAMSIZE;
	/* At least 1 byte is copied, the NUL */
	do {
		*--q = *--p;
	} while (p != params);
	ip = q;
	op = params;
	process_params();
#ifdef	DEBUG
	printf("Parameters: %s\n", p);
#endif
	/* Copy ramdisk to top of memory by longwords, tail first */
	if (seg[S_RAMDISK] != 0) {
		long			*dp, *sp;

		dp = (long *)(top_of_mem << 10);
		sp = (long *)((p = (char *)seg[S_RAMDISK]->p_paddr) +
			(i = seg[S_RAMDISK]->p_filesz));
		while (sp > (long *)p)
			*--dp = *--sp;
		printf("Ramdisk at %#X, size %#X\n",
			(setup->su_ramdisk_start = (unsigned long)dp),
			(setup->su_ramdisk_size = i));
	}
#ifdef	DEBUG
	printf("Ready\n");
	/* Delay so we can read display */
	for (i = 0; i < 0x7ffffff; i++)
		;
#endif
	xstart(seg[S_SETUP]->p_paddr);
	return (0);
}
